/***************************************************************************
 *                                                                         *
 *    File:         TeDSpell.cpp                                           *
 *                                                                         *
 *    Description:  Spell-checking manager of X-Window TeD                 *
 *                                                                         *
 *    Copyright:    (C) Copyright Paolo Galatola 1994.                     *
 *                  All rights reserved.                                   *
 *                                                                         *
 *    Date:         28 Nov 1996                                            *
 *                                                                         *
 ***************************************************************************/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <qlabel.h>                      // Labels
#include <qdialog.h>                     // Dialogs
#include <qframe.h>                      // Frames
#include <qpushbt.h>                     // Push buttons
#include <qchkbox.h>                     // Check boxes
#include <qradiobt.h>                    // Radio buttons
#include <qbttngrp.h>                    // Button groups
#include <qlined.h>                      // Line edit
#include <unistd.h>                      // Fork process
#include "TeDFrame.h"
#include "TeD.h"

#define SPELL_INIT1   0
#define SPELL_INIT2   1
#define SPELL_START   2
#define SPELL_READ    3
#define SPELL_PARSE   4
#define SPELL_SKIP    5
#define SPELL_CLOSE   6

// Spell-check window constructor

WinSpell::WinSpell(QWidget* parent, const char* name, WFlags f)
        : QDialog (0, name, TRUE, f)
{
  // Save parent
  hwnd = parent;

  // Sets frame & separator
  frame = new QFrame(this);
  separator = new QFrame(this);
  frame->setFrameStyle( QFrame::Panel | QFrame::Raised );
  separator->setFrameStyle( QFrame::HLine | QFrame::Sunken );

  // Sets labels
  Mlabel = new QLabel(this);
  Mlabel->setText("Replace: ");
  Mlabel->setAlignment(AlignCenter);
  Mlabel->setAutoResize(TRUE);
  Mlabelp = new QLabel(this);
  Mlabelp->setAlignment(AlignCenter);
  Mlabelp->setAutoResize(TRUE);

  // Sets line edit
  line = new QLineEdit(this);
  line->setMaxLength(ROWLEN);                  // Max ROWLEN characters

  // Sets combo box
  combo = new QComboBox(this);
  combo->setAutoResize(TRUE);

  // Sets buttons
  Replace = new QPushButton(this);
  Skip = new QPushButton(this);
  Accept = new QPushButton(this);
  Insert = new QPushButton(this);
  InsertLC = new QPushButton(this);
  Quit = new QPushButton(this);

  Replace->setText("Replace");
  Skip->setText("Skip");
  Accept->setText("Accept");
  Insert->setText("Insert");
  InsertLC->setText("Insert lc");
  Quit->setText("Quit");

  // Connect button signals
  connect(Replace, SIGNAL(clicked()), SLOT(replace_clicked()));
  connect(Skip, SIGNAL(clicked()), SLOT(skip_clicked()));
  connect(Accept, SIGNAL(clicked()), SLOT(accept_clicked()));
  connect(Insert, SIGNAL(clicked()), SLOT(insert_clicked()));
  connect(InsertLC, SIGNAL(clicked()), SLOT(insertlc_clicked()));
  connect(Quit, SIGNAL(clicked()), SLOT(quit_clicked()));

  // Connect combo box selection
  connect(combo, SIGNAL(activated(int)), SLOT(combo_update(int)));

  // Dim buttons combo box and line edit
  Replace->setEnabled(FALSE);
  Skip->setEnabled(FALSE);
  Accept->setEnabled(FALSE);
  Insert->setEnabled(FALSE);
  InsertLC->setEnabled(FALSE);
  combo->setEnabled(FALSE);
  line->setEnabled(FALSE);

  // Create timer for reading from ispell process
  rtimer = new QTimer(this);
  connect(rtimer, SIGNAL(timeout()), SLOT(read_ispell ()));

  // Create timer for writing to ispell process
  wtimer = new QTimer(this);
  connect(wtimer, SIGNAL(timeout()), SLOT(write_ispell ()));

  // Create timer for parsing input
  itimer = new QTimer(this);
  connect(itimer, SIGNAL(timeout()), SLOT(start_check ()));

  // Initialise variables
  command = SPELL_INIT1;
  spell_start = 1;

  // Start spell checker
  itimer->start(400, TRUE);

  // Resize window & show
  resize(365, 240);
  show();
}

// Spell-check window destructor

WinSpell::~WinSpell()
{
  // Free storage

  delete Mlabel;                 // Message box label
  delete Mlabelp;                // Message box progress label
  delete combo;                  // Combo box
  delete Replace;                // Replace button
  delete Skip;                   // Skip button
  delete Accept;                 // Accept button
  delete Insert;                 // Insert button
  delete InsertLC;               // Insert lower-case  button
  delete Quit;                   // Quit button
  delete separator;              // Separator line
  delete frame;                  // Frame
  delete rtimer;                 // Timer
  delete wtimer;                 // Timer
  delete itimer;                 // Timer
}


// Spell-check window resize event

void WinSpell::resizeEvent( QResizeEvent* )
{

  const int vspace = 10;           // Vertical space between buttons
  const int hspace = 30;           // Horizontal space between buttons
  const int hei = 25;              // Button height
  const int hlabel = 60;           // Label length
  int hpos;                        // Temporary storage
  int chei;                        // Combo box height
  int cwid;                        // Combo box width

  // Compute button length
  int blen = (this->width() - 4 * hspace) / 3;

  // Sets frame geometry
  frame->setGeometry(0, 0, this->width(), this->height());

  // Sets separator geometry
  hpos = this->height() - 3*hei - 4*vspace;
  separator->setGeometry(0, hpos, this->width(), 4);  

  // Sets labels geometry
  Mlabel->setGeometry(vspace, vspace, hlabel, hei);
  Mlabelp->setGeometry(vspace, hpos - hei - vspace,
		       this->width() - 2*vspace, hei);

  // Sets line edit geometry
  hpos = this->width() - 3*vspace - hlabel;
  line->setGeometry(2*vspace + hlabel, vspace, hpos, hei);

  // Sets combo box geometry
  cwid = combo->width() > blen ? combo->width() : blen;
  chei = combo->height() > hei ? combo->height() : hei;
  hpos = (this->width() - cwid) / 2;
  combo->setGeometry(hpos, 2*vspace + hei, cwid, chei);

  // Sets replace skip accept button geometry
  Replace->setGeometry (hspace, this->height() - 3*(vspace + hei), blen, hei);

  Skip->setGeometry (blen + 2*hspace,
                       this->height() - 3*(vspace + hei),blen, hei);

  Accept->setGeometry (this->width() - hspace - blen,
                       this->height() - 3*(vspace + hei), blen, hei);


  // Sets insert insert lower-case buttons geometry
  hpos = (this->width() - hspace) / 2 - blen;
  Insert->setGeometry (hpos, this->height() - 2*(vspace + hei), blen, hei);

  InsertLC->setGeometry (hpos + blen + hspace,
                         this->height() - 2*(vspace + hei), blen, hei);


  // Sets quit button geometry
  hpos = (this->width() - blen) / 2;
  Quit->setGeometry (hpos, this->height() - vspace - hei, blen, hei);

}


// Combo box selection

void WinSpell::combo_update(int i)
{

  // Insert selected hint in text window
   line->setText(combo->text(i));

}

// Replace button pressed

void WinSpell::replace_clicked()
{

  // Load contents of line editor in read buffer
  strcpy(read_buffer + 2, line->text());

  // Switch off cursor
  ((MainWindow *)hwnd)->Cursor->cursor_off();

  // Substitute marked text
  DelMark((MainWindow *)hwnd, read_buffer + 2);

  // Call skip button
  skip_clicked();
}


// Skip button pressed

void WinSpell::skip_clicked()
{

  // Dim buttons combo box and line edit
  Replace->setEnabled(FALSE);
  Skip->setEnabled(FALSE);
  Accept->setEnabled(FALSE);
  Insert->setEnabled(FALSE);
  InsertLC->setEnabled(FALSE);
  combo->setEnabled(FALSE);
  line->setEnabled(FALSE);

  // Update beginning of spell check
  extern unsigned int cursor_column;
  extern unsigned int cursor_line;
  extern unsigned int cursor_row;
  extern struct row *actual;

  column_check = cursor_column;
  rowline_check = cursor_row;
  line_check = cursor_line;
  row_check = actual;

  // Skip to beginning of unchecked text
  int i, length;
  length = strlen(read_buffer + 2);
  for (i = length; i > 0; --i)
    {
      ++column_check;
      if (column_check > strlen(row_check->text))
	  {
	    column_check = 0;
	    ++rowline_check;
	    if (row_check->carriage != cont)
	      ++line_check;
	    row_check = row_check->next;
	  }     
    }

  // Continue spell check
  itimer->start(1, TRUE);

}


// Accept button pressed

void WinSpell::accept_clicked()
{
  // Accept mispelled word as correct for this session
  strcpy(write_buffer,"@");
  strcat(write_buffer, read_buffer + 2);

  // Send command to ispell
  command = SPELL_SKIP;
  wtimer->start(1, TRUE);

}


// Insert button pressed

void WinSpell::insert_clicked()
{
  // Insert mispelled word in dictionary
  strcpy(write_buffer,"*");
  strcat(write_buffer, read_buffer + 2);

  // Send command to ispell
  command = SPELL_SKIP;
  wtimer->start(1, TRUE);


}

// Insert lower-case button pressed

void WinSpell::insertlc_clicked()
{
  // Insert lower-case mispelled word in dictionary
  strcpy(write_buffer,"&");
  strcat(write_buffer, read_buffer + 2);

  // Send command to ispell
  command = SPELL_SKIP;
  wtimer->start(1, TRUE);

}


// Quit button pressed

void WinSpell::quit_clicked ()
{
  // Save current personal dictionary
  strcpy(write_buffer, "#");
  command = SPELL_CLOSE;
  wtimer->start(1, TRUE);

}


// Main spell checker parser

void WinSpell::start_check()
{
  extern int outdes[2];

  switch(command) {

  case SPELL_INIT1:
    // Empty input buffer
    read(outdes[0], read_buffer, MAXBUFFER - 1);
    // Enter terse mode
    strcpy(write_buffer,"!");
    command = SPELL_INIT2;
    wtimer->start(1, TRUE);
    break;

  case SPELL_INIT2:
    // Enter TeX mode
    strcpy(write_buffer,"+");
    command = SPELL_START;
    wtimer->start(1, TRUE);
    break;

  case SPELL_START:
    // Start spelling check
    spell_in();

    // Run spell checker
    command = SPELL_READ;
    wtimer->start(1, TRUE);
    break;

  case SPELL_READ:
    // Read output from ispell
    command = SPELL_PARSE;
    rtimer->start(1, TRUE);
    break;

  case SPELL_PARSE:
    command = SPELL_START;
    spell_out();
    break;

  case SPELL_SKIP:
    command = SPELL_START;
    skip_clicked();
    break;

  case SPELL_CLOSE:
    // Close window
    close (FALSE);
  }
}


// Routine to read from ispell process

#include <errno.h>

void WinSpell::read_ispell()
{
  int rchar;                // Number of character read from pipe
  extern int outdes[2];
  extern int errno;

  // Read data from pipe
  rchar = read(outdes[0], read_buffer, MAXBUFFER - 1);
  if (rchar != -1) {
    // Terminates string and run main routine
    read_buffer[rchar-1] = (char) NULL;
    itimer->start(1, TRUE);
  }
  else {
    // An error occurred: check if input is not yet available
    if (errno == EAGAIN) {
      // Start read timer
      rtimer->start(1, TRUE);
    }
  }
}


// Routine to write to ispell process

void WinSpell::write_ispell()
{
  int wchar;                // Number of character to write to pipe
  extern int indes[2];
  extern int errno;

  // Add end of line
  wchar = strlen(write_buffer);
  write_buffer[wchar++] = (char) EOL;
  write_buffer[wchar] = (char) NULL;

  // Write data to pipe
  if (write(indes[1], write_buffer, wchar) == -1)
    {
    // An error occurred: check if output is not yet available
    if (errno == EAGAIN) {

      // Delete end of line and start write timer
      write_buffer[--wchar] = (char) NULL;
      wtimer->start(1, TRUE);
     }
    } 
  else
    {
    // Go to main parser
    itimer->start(1, TRUE);
    }
}


// Input parser

void WinSpell::spell_in()
{
#define BLANK 32

  if (spell_start)
    {
      // Find beginning of word to spell
      extern unsigned int cursor_column;
      extern unsigned int cursor_line;
      extern unsigned int cursor_row;
      extern struct row *actual;
      char prec;

      // Initialize data
      column_check = cursor_column;
      rowline_check = cursor_row;
      line_check = cursor_line;
      row_check = actual;

      do
	{
	if (column_check)
	  {
	  prec = *(row_check->text + column_check - 1);
	  if (prec != BLANK)
	    --column_check;
	  }
	else if (row_check->prev)
	  {
	  if (row_check->prev->carriage != cont)
	    prec = BLANK;
	  else
	    {
	    row_check = row_check->prev;
	    column_check = ROWLEN;
	    --rowline_check;
	    prec = *(row_check->text + column_check);
	    }
	  }
	else
	  prec = BLANK;
	} while (prec != BLANK);

      spell_start = 0;
    }

    // Initialize check buffer
    strcpy(write_buffer,"^");
    // Copy line to end
    strcat(write_buffer, row_check->text + column_check);

    // Concatenate continuing lines
    struct row *temp;
    temp = row_check;
    while (temp->carriage == cont)
      {
	temp = temp->next;
	if (strlen(write_buffer) + strlen(temp->text) < MAXBUFFER)
	  strcat(write_buffer, temp->text);
      }

    // Write status string
    char status[255];
    sprintf(status,"Checking line %d", line_check);
    Mlabelp->setText(status);
    Mlabelp->repaint(TRUE);
}


// Output parser

void WinSpell::spell_out()
{

  if (read_buffer[0] == '@')
    {
      // Non valid string: read again
      rtimer->start(1, TRUE);
    }

  else if (read_buffer[0] == '&' || read_buffer[0] == '#' )
    {
      int length;                           // Length of mispelled word
      int offset;                           // Offset of mispelled word
      int hints;                            // Number of hints
      int pos;                              // Position of string parsing
      int pos_end;                          // End of string insertion
      int i;                                // Counter

      // Mispelled word: compute length of mispelled word
      length = strcspn(read_buffer + 2, " ");
      pos = length + 3;

      // Break read_buffer at mispelled word for future use
      read_buffer[length + 2] = (char) NULL;

      // Compute number of hints and offset of mispelled word
      if (read_buffer[0] == '#')
	{
	  hints = 0;
	  sscanf(read_buffer + pos,"%d", &offset);
	}
      else
	{
	  sscanf(read_buffer + pos, "%d %d", &hints, &offset);
	  pos = pos + strcspn(read_buffer + pos, ":") + 2;
	}

      // Switch off cursor
      ((MainWindow *)hwnd)->Cursor->cursor_off();

      // Deselect text
      Deselect((MainWindow *)hwnd);

      // Move cursor to beginning of current line checked
      extern unsigned int cursor_column;
      extern unsigned int cursor_line;
      extern unsigned int cursor_row;
      extern struct row *actual;

      cursor_column = column_check;
      cursor_row = rowline_check;
      cursor_line = line_check;
      actual = row_check;

      // Move cursor to beginning of mispelled word
      for (i = offset; i > 1; --i)
	{
	  ++cursor_column;
	  if (cursor_column > strlen(actual->text))
	    {
	      cursor_column = 0;
	      ++cursor_row;
	      if (actual->carriage != cont)
		++cursor_line;
	      actual = actual->next;
	    }
	}

      // Move selection to end of mispelled word
      extern unsigned int mark_row_end;
      extern unsigned int mark_line_end;
      extern unsigned int mark_column_end;
      extern struct row *mark_pointer_end;

      mark_row_end = cursor_row;
      mark_line_end = cursor_line;
      mark_column_end = cursor_column;
      mark_pointer_end = actual;

      for (i = length; i > 0; --i)
	{
	  ++mark_column_end;
	  if (mark_column_end > strlen(mark_pointer_end->text))
	    {
	      mark_column_end = 0;
	      ++mark_row_end;
	      if (mark_pointer_end->carriage != cont)
		++mark_line_end;
	      mark_pointer_end = mark_pointer_end->next;
	    }
	}

      // Select mispelled word
      Select((MainWindow *)hwnd);
  
      // Centrate cursor
      CentrateCursor((MainWindow *)hwnd);

      // Insert hints in combo box
      combo->clear();
      for (i = 0; i < hints; ++i)
	{
	  pos_end = pos + strcspn(read_buffer + pos, ",");
	  if (i == hints - 1)
	    *strchr(read_buffer + pos, (char) EOL) = (char) NULL;
	  else
	    read_buffer[pos_end] = (char) NULL;
	  combo->insertItem(read_buffer + pos, -1);
	  pos = pos_end + 2;
	}

      // Insert first hint in text window
      if (hints)
	line->setText(combo->text(0));
      else
	{
	// Clear text window
	line->setText("");
	// Resize combo box
	combo->resize(Quit->width(), Quit->height());
	}

      // Enable buttons combo box and line edit
      Replace->setEnabled(TRUE);
      Skip->setEnabled(TRUE);
      Accept->setEnabled(TRUE);
      Insert->setEnabled(TRUE);
      InsertLC->setEnabled(TRUE);
      combo->setEnabled(TRUE);
      line->setEnabled(TRUE);
    }

  else
    {
      // No errors: skip continuation lines
      while (row_check->carriage == cont)
	{
	row_check = row_check->next;
	++rowline_check;
	}

      // Go on with next line
      if (row_check->next)
	{
	  row_check = row_check->next;
	  ++line_check;
	  ++rowline_check;
	  column_check = 0;
	  // Rerun command parser
	  itimer->start(1, TRUE);
	}
      else
	{
	  // Spell check done
	  Mlabelp->setText("Spell check done");
	  Mlabelp->repaint(TRUE);
	}
    }
}
